import copy
import datetime
import json
import os

import dbstream
import psycopg2
import re
import requests
from psycopg2.extras import RealDictCursor
from pyred.core.Column import change_columns_type, choose_columns_to_extend, columns_type_bool_to_str
from pyred.core.Table import create_table, create_columns
from pyred.core.tools.compare import schema_compare_tool
from pyred.core.tools.print_colors import C
import time


class RedDBStream(dbstream.DBStream):
    def __init__(self, instance_name, client_id):
        super().__init__(instance_name, client_id=client_id)
        self.instance_type_prefix = "RED"
        self.ssh_init_port = 6543

    def connection(self):
        connection_kwargs = self.credentials()
        try:
            con = psycopg2.connect(**connection_kwargs, cursor_factory=RealDictCursor)
        except psycopg2.OperationalError:
            time.sleep(2)
            if self.ssh_tunnel:
                self.ssh_tunnel.close()
                self.create_tunnel()
            con = psycopg2.connect(**connection_kwargs, cursor_factory=RealDictCursor)
        return con

    def _execute_query_custom(self, query):
        con = self.connection()
        cursor = con.cursor()
        try:
            cursor.execute(query)
        except Exception as e:
            cursor.close()
            con.close()
            raise e
        con.commit()
        try:
            result = cursor.fetchall()
        except psycopg2.ProgrammingError:
            result = None
        cursor.close()
        con.close()
        query_create_table = re.search("(?i)(?<=((create table ))).*(?= as)", query)
        if result:
            return [dict(r) for r in result]
        elif query_create_table:
            return {'execute_query': query_create_table}
        else:
            empty_list = []
            return empty_list

    def _send(self, data, replace, batch_size=1000):
        print(C.WARNING + "Initiate send_to_redshift..." + C.ENDC)
        con = self.connection()
        cursor = con.cursor()
        if replace:
            cleaning_request = '''DELETE FROM ''' + data["table_name"] + ''';'''
            print(C.WARNING + "Cleaning" + C.ENDC)
            self.execute_query(cleaning_request)
            print(C.OKGREEN + "[OK] Cleaning Done" + C.ENDC)

        boolean = True
        index = 0
        total_rows = len(data["rows"])
        total_nb_batches = len(data["rows"]) // batch_size + 1
        while boolean:
            temp_row = []
            for i in range(batch_size):
                if not data["rows"]:
                    boolean = False
                    continue
                temp_row.append(data["rows"].pop())

            final_data = []
            for x in temp_row:
                for y in x:
                    final_data.append(y)

            temp_string = ','.join(map(lambda a: '(' + ','.join(map(lambda b: '%s', a)) + ')', tuple(temp_row)))

            inserting_request = '''INSERT INTO ''' + data["table_name"] + ''' (\"''' + "\", \"".join(
                data["columns_name"]) + '''\") VALUES ''' + temp_string + ''';'''
            if final_data:
                try:
                    cursor.execute(inserting_request, final_data)
                except Exception as e:
                    cursor.close()
                    con.close()
                    raise e
            index = index + 1
            percent = round(index * 100 / total_nb_batches, 2)
            if percent < 100:
                print("\r   %s / %s (%s %%)" % (str(index), total_nb_batches, str(percent)), end='\r')
            else:
                print("\r   %s / %s (%s %%)" % (str(index), total_nb_batches, str(percent)))
        con.commit()

        cursor.close()
        con.close()
        print(C.HEADER + str(total_rows) + ' rows sent to Redshift' + C.ENDC)
        print(C.OKGREEN + "[OK] Sent to redshift" + C.ENDC)
        return 0

    def _send_data_custom(self,
                          data,
                          replace=True,
                          batch_size=1000,
                          other_table_to_update=None
                          ):
        """
        data = {
            "table_name" 	: 'name_of_the_redshift_schema' + '.' + 'name_of_the_redshift_table' #Must already exist,
            "columns_name" 	: [first_column_name,second_column_name,...,last_column_name],
            "rows"		: [[first_raw_value,second_raw_value,...,last_raw_value],...]
        }
        """
        data_copy = copy.deepcopy(data)
        try:
            self._send(data, replace=replace, batch_size=batch_size)
        except Exception as e:
            if "invalid input syntax for integer" in str(e).lower() \
                    or "invalid input syntax for type integer" in str(e).lower() \
                    or "invalid input syntax for type double precision" in str(e).lower() \
                    or "is out of range for type integer" in str(e).lower():
                change_columns_type(
                    self,
                    data=data_copy,
                    other_table_to_update=other_table_to_update
                )
            elif "invalid input syntax for type boolean:" in str(e).lower():
                columns_type_bool_to_str(
                    self,
                    data=data_copy,
                    other_table_to_update=other_table_to_update
                )
            elif "value too long for type character" in str(e).lower():
                choose_columns_to_extend(
                    self,
                    data=data_copy,
                    other_table_to_update=other_table_to_update
                )
            elif "does not exist" in str(e).lower() and "column" in str(e).lower():
                create_columns(
                    self,
                    data=data_copy,
                    other_table_to_update=other_table_to_update
                )
            elif "does not exist" in str(e).lower() and ("relation" in str(e).lower() or "schema" in str(e).lower()):
                print("Destination table doesn't exist! Will be created")
                create_table(
                    self,
                    data=data_copy,
                    other_table_to_update=other_table_to_update
                )
                replace = False

            else:
                raise e

            self._send_data_custom(data_copy, replace=replace, batch_size=batch_size,
                                   other_table_to_update=other_table_to_update)

    def clean(self, selecting_id, schema_prefix, table):
        print('trying to clean table %s.%s using %s' % (schema_prefix, table, selecting_id))
        cleaning_query = """
                DELETE FROM %(schema_name)s.%(table_name)s WHERE %(id)s IN (SELECT distinct %(id)s FROM %(schema_name)s.%(table_name)s_temp);
                INSERT INTO %(schema_name)s.%(table_name)s 
                SELECT * FROM %(schema_name)s.%(table_name)s_temp;
                DELETE FROM %(schema_name)s.%(table_name)s_temp;
                """ % {"table_name": table,
                       "schema_name": schema_prefix,
                       "id": selecting_id}
        self.execute_query(cleaning_query)
        print('cleaned')

    def get_max(self, schema, table, field, filter_clause=""):
        try:
            r = self.execute_query("SELECT max(%s) as max FROM %s.%s %s" % (field, schema, table, filter_clause))
            return r[0]["max"]
        except IndexError:
            return None
        except (psycopg2.ProgrammingError, psycopg2.errors.InvalidSchemaName) as e:
            if "relation" in str(e) or "schema" in str(e):
                return None
            raise e

    def get_data_type(self, table_name, schema_name):
        query = """ SELECT
                column_name, data_type
                FROM
                information_schema.columns
                WHERE
                table_name = '%s' and table_schema = '%s'
                union 
                select col_name as column_name , col_type as data_type
                from pg_get_late_binding_view_cols() cols(view_schema name, view_name name, col_name name, col_type varchar, col_num int)
                WHERE
                view_name = '%s' and view_schema = '%s'
                
                """ % (table_name, schema_name, table_name, schema_name)

        return self.execute_query(query=query)

    def create_view_from_columns(self, view_name, columns, schema_name, table_name):
        view_query = '''DROP VIEW IF EXISTS %s ;CREATE VIEW %s as (SELECT %s FROM %s.%s)''' \
                     % (view_name, view_name, columns, schema_name, table_name)
        self.execute_query(view_query)

    def create_schema(self, schema_name):
        self.execute_query("CREATE SCHEMA %s" % schema_name)

    def drop_schema(self, schema_name):
        self.execute_query("DROP SCHEMA %s CASCADE" % schema_name)

    def schema_compare(self, schema_ref, new_schema_test):
        schema_compare_tool(self, schema_ref, new_schema_test)



import pandas as pd

def schema_compare_tool(self, schema_ref, new_schema_test):
    # same table uploaded ?
    query1 = """select "table_name" from information_schema.tables where table_schema='%s'""" % schema_ref
    query2 = """select "table_name" from information_schema.tables where table_schema='%s'""" % new_schema_test
    table_schema_ref = pd.DataFrame(self.execute_query(query1))
    table_new_schema_test = pd.DataFrame(self.execute_query(query2))
    if table_schema_ref['table_name'].equals(table_new_schema_test['table_name']):
        print("Same tables in both schemas")
    if not table_schema_ref['table_name'].equals(table_new_schema_test['table_name']):
        print("Table not loaded in schema %s:" % new_schema_test)
        df = table_schema_ref.merge(table_new_schema_test, how='outer', indicator=True).loc[lambda x: x['_merge'] == 'left_only']
        if not df.empty:
            print(df)
        print("")

    # for each data loaded : same columns loaded ? if loaded same type? if same type : equals ?
    for table in table_new_schema_test['table_name']:
        print("")
        print('\033[1m' + table + '\033[0m')
        query1 = """select "column_name" from information_schema.columns where table_schema='%s' and table_name='%s'""" \
                 % (schema_ref, table)
        query2 = """select "column_name" from information_schema.columns where table_schema='%s' and table_name='%s'""" \
                 % (new_schema_test, table)
        columns_table_1 = pd.DataFrame(self.execute_query(query1))
        columns_table_2 = pd.DataFrame(self.execute_query(query2))
        if columns_table_1['column_name'].equals(columns_table_2['column_name']):
            print("     Same columns in both tables")
        if not columns_table_1['column_name'].equals(columns_table_2['column_name']):
            print("     columns not loaded :")
            df = columns_table_1.merge(columns_table_2, how='outer', indicator=True).loc[
                lambda x: x['_merge'] == 'left_only']
            if not df.empty:
                print(df)
            print("     columns loaded in addition:")
            df = columns_table_1.merge(columns_table_2, how='outer', indicator=True).loc[
                lambda x: x['_merge'] == 'right_only']
            if not df.empty:
                print(df)
        for column in columns_table_1.merge(columns_table_2, how='inner', indicator=True)['column_name']:
            print("         ", column)
            query1 = """select "udt_name" from information_schema.columns where table_schema='%s' and table_name='%s' and column_name='%s'""" \
                     % (schema_ref, table, column)
            query2 = """select "udt_name" from information_schema.columns where table_schema='%s' and table_name='%s' and column_name='%s'""" \
                     % (new_schema_test, table, column)
            type1 = str(self.execute_query(query1))
            type2 = str(self.execute_query(query2))
            if type1 != type2:
                print("         ", "type from schema 1 :", type1, "// type from schema 2:", type2)
            if type1 == type2:
                query1 = """select '%s' from %s.%s""" \
                         % (column, schema_ref, table)
                query2 = """select '%s' from %s.%s""" \
                         % (column, new_schema_test, table)
                value1 = pd.DataFrame(self.execute_query(query1))
                value2 = pd.DataFrame(self.execute_query(query2))
                print("         ", value2.equals(value1))


import copy
import datetime
import pandas as pd
import psycopg2

def change_type(_dbstream, table_name, column_name, type):
    query = """
    ALTER TABLE %(table_name)s ADD COLUMN %(new_column_name)s %(type)s;
    UPDATE %(table_name)s SET  %(new_column_name)s = %(column_name)s;
    ALTER TABLE %(table_name)s DROP COLUMN %(column_name)s CASCADE ;
    ALTER TABLE %(table_name)s RENAME COLUMN %(new_column_name)s TO %(column_name)s;
    """ % {
        "table_name": table_name,
        "column_name": column_name,
        "new_column_name": column_name + "_new",
        "type": type
    }
    _dbstream.execute_query(query)
    return query

def bool_to_str(_dbstream, table_name, column_name):
    query = """
    ALTER TABLE %(table_name)s ADD COLUMN %(new_column_name)s VARCHAR(255);
    UPDATE %(table_name)s SET  %(new_column_name)s = case when %(column_name)s=True then 'True' when %(column_name)s=False then 'False' end;
    ALTER TABLE %(table_name)s DROP COLUMN %(column_name)s CASCADE ;
    ALTER TABLE %(table_name)s RENAME COLUMN %(new_column_name)s TO %(column_name)s;
    """ % {
        "table_name": table_name,
        "column_name": column_name,
        "new_column_name": column_name + "_new"
    }
    _dbstream.execute_query(query)
    return query

def get_columns_length(_dbstream, schema_name, table_name):
    query = """
    SELECT column_name, character_maximum_length
    FROM information_schema.columns
    WHERE table_name='%s' and table_schema='%s'
    AND character_maximum_length IS NOT NULL  
    """ % (table_name, schema_name)
    d = {}
    r = _dbstream.execute_query(query)
    for i in r:
        d[i["column_name"]] = i["character_maximum_length"]
    return d

def get_columns_type(_dbstream, schema_name, table_name):
    query = """
    SELECT column_name, udt_name
    FROM information_schema.columns
    WHERE table_name='%s' and table_schema='%s' 
    """ % (table_name, schema_name)
    d = {}
    r = _dbstream.execute_query(query)
    for i in r:
        d[i["column_name"]] = i["udt_name"]
    return d


def choose_columns_to_extend(_dbstream, data, other_table_to_update):
    table_name = data["table_name"].split('.')
    columns_length = get_columns_length(_dbstream, table_name=table_name[1], schema_name=table_name[0])
    rows = data["rows"]
    columns_name = data["columns_name"]
    df = pd.DataFrame(rows, columns=columns_name)

    for c in columns_name:
        example = find_sample_value(df, c, columns_name.index(c))[0]
        if isinstance(example, str):
            if len(str(example.encode())) >= 255:
                if not columns_length.get(c) or columns_length.get(c) < len(str(example.encode())):
                    change_type(_dbstream, table_name=data["table_name"], column_name=c, type="VARCHAR(65000)")
                    if other_table_to_update:
                        change_type(_dbstream, table_name=other_table_to_update, column_name=c, type="VARCHAR(65000)")

def change_columns_type(_dbstream, data, other_table_to_update):
    table_name = data["table_name"].split('.')
    columns_type = get_columns_type(_dbstream, table_name=table_name[1], schema_name=table_name[0])
    rows = data["rows"]
    columns_name = data["columns_name"]
    df = pd.DataFrame(rows, columns=columns_name)

    for c in columns_name:
        example = find_sample_value(df, c, columns_name.index(c))[0]
        if isinstance(example, float):
            if columns_type.get(c) != "float8":
                change_type(_dbstream, table_name=data["table_name"], column_name=c, type="float8")
                if other_table_to_update:
                    change_type(_dbstream, table_name=other_table_to_update, column_name=c, type="float8")
        if isinstance(example, str):
            if columns_type.get(c) not in ["varchar", "bool", "timestamp"]:
                change_type(_dbstream, table_name=data["table_name"], column_name=c, type="VARCHAR(255)")
                if other_table_to_update:
                    change_type(_dbstream, table_name=other_table_to_update, column_name=c, type="VARCHAR(255)")
        if isinstance(example, int) and (example > 2147483646):
            if columns_type.get(c) != "int8":
                change_type(_dbstream, table_name=data["table_name"], column_name=c, type="int8")
                if other_table_to_update:
                    change_type(_dbstream, table_name=other_table_to_update, column_name=c, type="float8")

def columns_type_bool_to_str(_dbstream, data, other_table_to_update):
    table_name = data["table_name"].split('.')
    columns_type = get_columns_type(_dbstream, table_name=table_name[1], schema_name=table_name[0])
    rows = data["rows"]
    columns_name = data["columns_name"]
    df = pd.DataFrame(rows, columns=columns_name)

    for c in columns_name:
        example = find_sample_value(df, c, columns_name.index(c))[0]
        if isinstance(example, str):
            if columns_type.get(c) == "bool":
                bool_to_str(_dbstream, table_name=data["table_name"], column_name=c)
                if other_table_to_update:
                    bool_to_str(_dbstream, table_name=other_table_to_update, column_name=c)

def detect_type(_dbstream, name, example):
    print('Define type of %s...' % name)
    try:
        query = "SELECT CAST('%s' as TIMESTAMP)" % example
        _dbstream.execute_query(query)
        return "TIMESTAMP"

    except psycopg2.Error:
        pass
    if isinstance(example, datetime.date):
        return "TIMESTAMP"
    elif isinstance(example, str):
        if len(example) >= 255:
            return "VARCHAR(65000)"
        return "VARCHAR(255)"
    elif isinstance(example, bool):
        return "BOOLEAN"
    elif isinstance(example, int):
        if example > 2147483646:
            return "BIGINT"
        else:
            return "INTEGER"
    elif isinstance(example, float):
        return "FLOAT"
    else:
        return "VARCHAR(255)"

def convert_to_bool(x):
    if x.lower() == "true" or x == 1 or x.lower() == "t":
        return True
    if x.lower() == "false" or x == 0 or x.lower() == "f":
        return False
    else:
        raise Exception

def convert_to_int(x):
    if x[-2:] == ".0":
        return int(x.replace(".0",""))
    else:
        return int(x)

def len_or_max(s):
    if isinstance(s, str):
        return len(s)
    return s

def find_sample_value(df, name, i):
    df1 = df[name].dropna()
    try:
        df1 = df1.apply(lambda x: str(x))
    except:
        pass
    try:
        df1 = df1.apply(lambda x: convert_to_bool(x))
    except:
        try:
            df1 = df1.apply(lambda x: convert_to_int(x))
        except:
            try:
                df1 = df1.apply(lambda x: float(x))
            except:
                pass
    df1_copy = copy.deepcopy(df1)
    if df1.dtype == 'object':
        df1 = df1.apply(lambda x: (str(x.encode()) if isinstance(x, str) else x) if x is not None else '')
        if df1.empty:
            return None, None
        else:
            return df1_copy[df1.map(len_or_max) == df1.map(len_or_max).max()].iloc[0], df1_copy[df1.map(len_or_max) == df1.map(len_or_max).min()].iloc[0]
    elif df1.dtype == 'int64':
        max = int(df1.max())
        min = int(df1.min())
        return max, min
    elif df1.dtype == 'float64':
        max = float(df1.max())
        min = float(df1.min())
        return max, min
    else:
        rows = df.values.tolist()
        for row in rows:
            if row[i] is not None:
                return row[i], row[i]
        return None, None
